import os
import sys
import shutil
#import subprocess
import argparse

import pwd
import json
import ssl
import urllib.request
import urllib.error
from pathlib import Path

from platformdirs import (
    user_data_dir, 
    user_runtime_dir,
    user_config_dir,
    user_log_dir,
)

#import sentry_sdk
#sentry_sdk.init(
#    dsn="https://bbacdfaf17304b809e02b7ab39a64226@sentry.mirimus.com/17",
#    traces_sample_rate=1.0
#)

#from contextlib import contextmanager
#from twitter.common.contextutil import pushd, temporary_dir

pwuid = pwd.getpwuid(os.getuid())
current_user = pwuid[0]

data_dir = Path(user_data_dir("vconf", current_user))
data_dir.mkdir(parents=True, exist_ok=True)

plugins_dir = data_dir / "plugins"
plugins_dir.mkdir(parents=True, exist_ok=True)

run_dir = Path(user_runtime_dir("vconf", current_user))
run_dir.mkdir(parents=True, exist_ok=True)

log_dir = Path(user_log_dir("vconf", current_user))
log_dir.mkdir(parents=True, exist_ok=True)

config_dir = Path(user_config_dir("vconf", current_user))
config_dir.mkdir(parents=True, exist_ok=True)

#venv_dir = Path(data_dir) / '.venv'
env_file = Path(".env")

parser = argparse.ArgumentParser(
    prog="vconf",
    description="This is VConf ...",
    epilog="Thanks for using %(prog)s! :)",
    #argument_default=argparse.SUPPRESS,
)

parser.add_argument("-c", "--clean", action="store_true", default=False, help="clean installation")
parser.add_argument("-l", "--localdev", action="store_true", default=False, help="local dev")
parser.add_argument("-U", "--pip-upgrade-deps", action="store_true", default=True, help="upgrade pip dependencies")
parser.add_argument("-a", "--api-root", action="store", default="https://vc.eloquentbits.com", help="VConf Cloud API Root")
parser.add_argument("-d", "--device-id", action="store", default="", help="VConf Device ID")
parser.add_argument("-t", "--secret-key", action="store", default="", help="VConf Secret Key")

ns = parser.parse_args()

#py = Path("/usr/bin/python")
#if not py.exists():
#    py = Path("/usr/bin/python3")

#if not py.exists():
#    parser.exit(1, message=f"unable to locate a Python interpreter.")

#if not (venv_dir.exists() and venv_dir.is_dir()):
#    print(f'creating Python venv @ {venv_dir.absolute()}')
#    subprocess.check_call([py, "-m", "venv", str(venv_dir.absolute())])


#if ns.pip_upgrade_deps:
#    subprocess.check_call([
#        str(venv_dir / Path("bin/python")), 
#        "-m",
#        "pip",
#        "install", 
#        "--require-virtualenv",
#        "sseclient-py", 
#        "requests",
#        "pyzmq",
#        "sentry-sdk",
#        "python-dotenv",
#    ])

#    editable = ""
#    if ns.localdev == True:
#        editable = "--editable"
#        vconf_client_utils_path = Path(input("provide path to vconf_client_utils repo: "))
#        if not vconf_client_utils_path.exists():
#            parser.exit(1, message=f"{vconf_client_utils_path} does not exist. cannot install package in editable mode")
#        vconf_client_utils_path  = vconf_client_utils_path.absolute()
#    else:
#        # FIXME
#        vconf_client_utils_path = "https://pypi.vc.eloquentbits.com/packages/vconf_client_utils-0.1.dev23+gffe51b0.tar.gz\#md5\=9056681f4ead2f2e69c6236111f343c7"
#
#    subprocess.check_call([
#        (venv_dir / Path("bin/python")).absolute(), 
#        (venv_dir / Path("bin/pip")).absolute(), 
#        "install", 
#        "--require-virtualenv",
#        editable,
#        vconf_client_utils_path,
#    ])

def get_var_by_name(lines, var_name):
    try:
        line = list(filter(lambda x: x.split("=")[0] == var_name, lines))[0]
        return line.split("=")[1].strip()
    except IndexError:
        return

def query_vconf_cloud(device_id, secret_key, api_root):
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    url = f"{api_root}/env-bootstrap/{device_id}/"
    try:
        headers={
            "Authorization": f"Token {secret_key}",
            "X-VCONF-CLIENT-USER": f"{current_user}:{pwuid.pw_uid}:{pwuid.pw_gid}",
            "X-VCONF-DATA-DIR": str(data_dir),
            "X-VCONF-RUN-DIR": str(run_dir),
            "X-VCONF-CONFIG-DIR": str(config_dir),
            "X-VCONF-PLUGINS-DIR": str(plugins_dir),
            "X-VCONF-LOG-DIR": str(log_dir),
        }
        req = urllib.request.Request(
            url=url, 
            headers=headers,
        )
        #for dir in ["data", "run", "config", "plugins", "log"]:
        #    path = headers.get(f'X-VCONF-{dir.upper()}-DIR')
        #    print(f"{dir} dir: {path}")

        with urllib.request.urlopen(req, context=ctx) as response:
            if response.code == 403:
                parser.exit(1, message=f"Request Forbidden with secret_key {secret_key} @ {url}")

            body = response.read().decode()
            data = json.loads(body)
            with open(env_file, "w") as f:
                f.truncate()
                f.write(f"VCONF_DEVICE_ID={device_id}\n")
                f.write(f"VCONF_SECRET_KEY={secret_key}\n")
                f.write(f"VCONF_ADDR={api_root}\n")
                f.write(f"VCONF_DATA_DIR={data_dir}\n")
                f.write(f"VCONF_RUN_DIR={run_dir}\n")
                f.write(f"VCONF_CONFIG_DIR={config_dir}\n")
                f.write(f"VCONF_PLUGINS_DIR={plugins_dir}\n")
                f.write(f"VCONF_LOG_DIR={log_dir}\n")
                #f.write(f"VCONF_STATE_DIR={state_dir}\n")
                for k, v in data.get("settings", {}).items():
                    f.write(f"{k}={v}\n")
                    #print(f"[bootstrap-venv] writing: {k}={v}\n")
                    #uwsgi.cache_update(k, v, 0, uwsgi_cache_name)
    except urllib.error.HTTPError as e:
    #    uwsgi.log(fh, e.code, url)
        print(f"could not connect to: {url}. {str(e)}")
        parser.exit(1, message=f"unable to connect to VConf Cloud.")
        #HTTP Error 502: Bad Gateway

    #except urllib.error.URLError as e:
    #    if hasattr(e, 'reason'):
    #        log(fh, e.reason, url)
    #    elif hasattr(e, 'code'):
    #        log(fh, e.code, url)

def gather_settings(device_id, secret_key, api_root):
    print(f"gathering settings. {api_root=}")
    if env_file.exists():
        with open(env_file, "r") as f:
            lines = f.readlines()
            if not device_id:
                device_id = get_var_by_name(lines, "VCONF_DEVICE_ID")
            if not secret_key:
                secret_key = get_var_by_name(lines, "VCONF_SECRET_KEY")
            if not api_root:
                api_root = get_var_by_name(lines, "VCONF_ADDR")

    if not device_id:
        device_id = input("Please enter your VConf Device ID: ")
    if not secret_key:
        secret_key = input("Please enter your VConf Secret Key: ")

    if not all([device_id, secret_key, api_root]):
        print(f"{device_id=} {secret_key=} {api_root=}")
        parser.exit(1, message=f"could not gather all the required settings.")

    query_vconf_cloud(device_id, secret_key, api_root)

def install_pip_pyz():
    pip_pyz =  Path(data_dir) / "pip.pyz"
    if not pip_pyz.exists():
        print(f"[bootstrap-venv] downloading {str(pip_pyz.absolute())}")
        with urllib.request.urlopen("https://bootstrap.pypa.io/pip/pip.pyz") as response, \
                open(pip_pyz, 'wb') as out_file:
            shutil.copyfileobj(response, out_file)

def clean():
    print(f"removing Python venv @ {venv_dir.absolute()}")
    shutil.rmtree(venv_dir)


#UWSGI_BINARY_PATH = 'vconf.so'
#UWSGI_BOOTSTRAP_PATH = 'bootstrap_pex.py'
#UWSGI_RESOURCE_PATH = 'vconf.resources'

#def write_resource(filename, resource_path=UWSGI_RESOURCE_PATH):
#    print(f"[vconf-main] {filename=} {resource_path=}")
    #with open(filename, 'wb') as fh:
        #import pkg_resources
        #for chunk in pkg_resources.resource_stream(resource_path, filename):
        #    fh.write(chunk)

#@contextmanager
#def uwsgi_context(*args, **kwargs):
#  """A context manager that provides a temporary directory with a pex-bootstrapped uwsgi.

#     Usage:

#       import os
#       from pyuwsgi import UWSGI_BINARY_PATH, uwsgi_context

#       with uwsgi_context() as uwsgi_path:
#         uwsgi_bin_path = os.path.join(uwsgi_path, UWSGI_BINARY_PATH)
#         p = subprocess.Popen([uwsgi_bin_path, '--arg1', '--arg2'])
#         ...


#       >>> with uwsgi_context() as uwsgi_path:
#       ...   print uwsgi_path
#       ...   os.listdir(uwsgi_path)
#       ...
#       /private/var/folders/3t/xkwqrkld4xxgklk2s4n41jb80000gn/T/tmpCY4JhN
#       ['pex_uwsgi']

#  """
#  old_cwd = os.getcwd()
#  with temporary_dir(*args, **kwargs) as temp_dir:
#    with pushd(temp_dir):
#      write_resource(UWSGI_BINARY_PATH)
#      write_resource(UWSGI_BOOTSTRAP_PATH)
#      with pushd(old_cwd):
#        yield temp_dir
#

#def activate_pex():
#  sys.stderr.write('[pex_uwsgi] bootstrapping..\n')

#  entry_point = os.environ.get('UWSGI_PEX')
#  if not entry_point:
#    sys.stderr.write(
#      '[pex_uwsgi] couldnt determine pex from UWSGI_PEX environment variable, bailing!\n'
#    )
#    sys.exit(1)

#  sys.stderr.write('[pex_uwsgi] entry_point=%s\n' % entry_point)

#  sys.path[0] = os.path.abspath(sys.path[0])
#  sys.path.insert(0, entry_point)
#  sys.path.insert(0, os.path.abspath(os.path.join(entry_point, '.bootstrap')))

#  from _pex import pex_bootstrapper

#  if hasattr(pex_bootstrapper, 'bootstrap_pex_env'):
#    # New-style pex bootstrapping.
#    pex_bootstrapper.bootstrap_pex_env(entry_point)
#  else:
    # Old and even older-style pex bootstrapping.
#    from _pex.environment import PEXEnvironment
#    from _pex.pex_info import PexInfo

#    try:
#      pex_bootstrapper.monkeypatch_build_zipmanifest()
#    except AttributeError:
#      pass

#    try:
#      from _pex.finders import register_finders
#      register_finders()
#    except (ImportError, NameError):
#      pass

#    pex_info = PexInfo.from_pex(entry_point)
#    env = PEXEnvironment(entry_point, pex_info)
#    env.activate()

  # import & reload in our newly bootstrapped environment
#  import pkg_resources
#  reload(pkg_resources)

#  sys.stderr.write('[pex_uwsgi] sys.path=%s\n\n' % sys.path)
#  return

#activate_pex()

def init():
    print(f"=========== VConf init ===================")

    #import sysconfig
    #print(f"{sysconfig.get_python_inc()=}")
    #print(f"{sysconfig.get_python_inc(plat_specific=True)=}")

    #with uwsgi_context() as uwsgi_path:
    #    uwsgi_bin_path = os.path.join(uwsgi_path, UWSGI_BINARY_PATH)
    #    print(f"{uwsgi_path=} {uwsgi_bin_path=}")
        #print(os.listdir(uwsgi_path))

    if ns.clean:
        print("cleaning installation")
        clean()
        return

    device_id = ns.device_id or os.environ.get("VCONF_DEVICE_ID")
    secret_key = ns.secret_key or os.environ.get("VCONF_SECRET_KEY")
    api_root = ns.api_root or os.environ.get("VCONF_ADDR")

    gather_settings(device_id, secret_key, api_root)
    #install_pip_pyz()

    emperor_wrapper = f"/home/{current_user}/.local/bin/uwsgi"
    args = [
        #'--binary-path=vconf',
        "--master",
        #"--strict",
        "--show-config",
        "--vacuum",
        "--workers=1",
        "--threads=3",
        #"--mules=3",
        "--mule",
        "--mule",
        "--mule",

        f"--uid={pwuid.pw_uid}",
        f"--gid={pwuid.pw_gid}",

        f"--emperor-wrapper={emperor_wrapper}",

        #set-ph=localdev=true
        "--set-ph=vconf_service_name=Emperor",
        #set-ph = vconf_device_id=12

        #"--exec-asap = print: running with localdev=%(localdev)",
        #"--log-encoder=format %[[33m${msgnl}%[[0m",
        #"plugin-dir = plugins",
        f"--plugin-dir={data_dir}/plugins",

        #plugin = vconf_provision_dirs
        #vconf-user-name = %U
        #vconf-user-id = %u
        "--cache2=name=vconf-settings,items=100",
        "--auto-procname",
        "--emperor-procname=VConf %(vconf_service_name):",
        "--procname-master=VConf %(vconf_service_name) Master",
        "--procname-prefix-spaced=VConf %(vconf_service_name):",
        "--emperor-nofollow",
        # TODO
        #exec-as-user-atexit =  hc ping fail service
        f"--virtualenv={data_dir}/.venv",
        #pythonpath=/home/pk/.local/share/vconf/.venv/lib/python3.11/site-packages
        "--import=vconf_client_utils.bootstrap.bootstrap",

        f"--master-fifo={run_dir}/vconf-master-fifo.sock",

        #"--emperor-stats-server=127.0.0.1:5003",
        f"--emperor-stats-server={run_dir}/vconf-stats.sock",

        f"--pidfile={run_dir}/vconf.pid",
        f"--socket={run_dir}/vconf.sock",

        #smart-attach-daemon = $(HOME)/run/redis-server.pid /usr/bin/redis-server $(HOME)/etc/redis.conf
        #smart-attach-daemon = %(vconf_run_dir)/nginx.pid /home/%U/.local/bin/nginx
        #smart-attach-daemon = %(vconf_run_dir)/dnsmasq.pid /home/%U/.local/bin/dnsmasq -C ~/.config/vconf/dnsmasq/dnsmasq.conf -k

        #privileged-binary-patch-arg = nginx -g "daemon off;"
        #add-cache-item = CACHED_RDS_USER=thisIsARedisUser

        #ini = :zeromq-imperial-monitor
        "--emperor=zmq://tcp://127.0.0.1:5250",
        #ini = :zeromq-imperial-publish
        "--import=vconf_client_utils.daemons.emperor_zeromq_publish",

        #ini = :emperor-sse
        "--import=vconf_client_utils.daemons.sse_client",

        #ini = :emperor-fastrouter
        #fastrouter = %(vconf_run_dir)/FastRouter.sock
        "--fastrouter=127.0.0.1:3034",
        #exec-as-user = chgrp www-data %(vconf_run_dir)/FastRouter.sock

        f"--fastrouter-subscription-server={run_dir}/SubscriptionServer.sock",
        f"--fastrouter-stats-server={run_dir}/FastRouter-stats.sock",
        "--fastrouter-cheap",
        "--fastrouter-post-buffering=8192",

        # enable the notification socket
        f"--notify-socket={run_dir}/notify.socket",
        # pass it in subscriptions
        f"--subscription-notify-socket={run_dir}/notify.socket",

        "--import=vconf_client_utils.launch_mules",
    ]


    # we import `vconf` with `dlopen` flags set to `os.RTLD_GLOBAL` such that
    # uwsgi plugins can discover uwsgi's globals (notably `extern ... uwsgi`)
    #if hasattr(sys, 'setdlopenflags'):
    #    orig = sys.getdlopenflags()
    #    try:

            #if 'LD_LIBRARY_PATH' not in os.environ:
            #    path = Path(__file__).parent.resolve()
            #    os.environ['LD_LIBRARY_PATH'] = f"{path}/vconf.libs/"
            #print(f"{os.environ['LD_LIBRARY_PATH']=}")
    #        sys.setdlopenflags(orig | os.RTLD_GLOBAL)
    #        import vconf
    #    finally:
    #        sys.setdlopenflags(orig)
    #else:  # ah well, can't control how dlopen works here
    #    import vconf

    import ctypes
    #'/home/pk/dev/vconf-test-non-root/venv/bin/vconf.so'
    vconf_so = "vconf-0.0.2.data/scripts/vconf.so" #"vconf.so"
    vconf = ctypes.CDLL(vconf_so, mode=ctypes.RTLD_GLOBAL)
    print(vconf)

    # build command line args
    argv = (ctypes.c_char_p * (len(args)+1))()
    for pos, arg in enumerate(args):
        argv[pos] = arg.encode()
    # inform the uwsgi engine, the passed environ is not safe to overwrite
    envs = (ctypes.c_char_p * 1)()

    print(f"{args=}")
    print(f"{argv=}")
    print(f"{envs=}")

    vconf.uwsgi_init(len(args), argv, envs)

#run = vconf.run
run = init



"""

PYUWSGI_ARGS = [
    "--master",
    "--strict",
    "--need-app",
    "--module",
    ":".join(WSGI_APPLICATION.rsplit(".", 1)),
    "--no-orphans",
    "--vacuum",
    "--auto-procname",
    "--enable-threads",
    "--offload-threads=4",
    "--thunder-lock",
    "--static-map",
    "=".join([STATIC_URL.rstrip("/"), STATIC_ROOT]),
    "--static-expires",
    "/* 7776000",
]



uwsgi.cache_update("VCONF_SECRET_KEY", vc_secret_key, 0, uwsgi_cache_name)
uwsgi.cache_update("VCONF_DEVICE_ID", vc_device_id, 0, uwsgi_cache_name)
uwsgi.cache_update("VCONF_ADDR", vc_addr, 0, uwsgi_cache_name)

cached_rds_user = uwsgi.cache_get("CACHED_RDS_USER", uwsgi_cache_name)
if cached_rds_user:
    uwsgi.log(f"[bootstrap-venv] {cached_rds_user.decode()}")
else:
    uwsgi.log(f"[bootstrap-venv] NO CACHED_RDS_USER found in cache!!")
uwsgi.cache_set("CACHED_RDS_USER", "thisIsARedisUser", 0, uwsgi_cache_name)

venv_dir = Path(data_dir) / '.venv'
if not (venv_dir.exists() and venv_dir.is_dir()):
    uwsgi.log(f'[bootstrap-venv] virtualenv not set up @ {str(venv_dir.absolute())}')
    #uwsgi.log(f'[bootstrap-venv] creating virtualenv @ {str(venv_dir.absolute())}')
    uwsgi.stop()
    #from microvenv import create 
    #create(venv_dir)

#uwsgi_cache_name = "vconf-settings"
#vc_device_id = os.environ.get("VCONF_DEVICE_ID")
#vc_secret_key = os.environ.get("VCONF_SECRET_KEY")
#vc_addr = os.environ.get("VCONF_ADDR")
#uwsgi.cache_update("VCONF_RUN_DIR", str(run_dir.absolute()), 0, uwsgi_cache_name)
"""

